/* Copyright (C) 2001 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1999 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1996 DJ Delorie, see COPYING.DJ for details */
/* Copyright (C) 1995 DJ Delorie, see COPYING.DJ for details */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <limits.h>
#include <assert.h>
#include <errno.h>
#if defined(__DJGPP__)
#include <unistd.h>
#include <io.h>
#include <libc/dosio.h>
#include <go32.h>
#include <dpmi.h>
#define tbsize _go32_info_block.size_of_transfer_buffer
#elif defined(_MSC_VER)
#include <io.h>
#else
#include <unistd.h>
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif
#ifndef PATH_MAX
#define PATH_MAX 512
#endif

/* options */
int opt_destubify = 0;
int opt_generate = 0;
char opt_stubfile[PATH_MAX+1] = { 0 };
int opt_verbose = 0;
#define v_printf if (opt_verbose) printf

/* default stub */
static const unsigned char stub_bytes[] = {
#include "stub.h"
};

unsigned long get32(const unsigned char *ptr)
{
/* avoid sign problems as `unsigned char' promotes to `int' */
#define P(i, n) (((unsigned long) ptr[i]) << n)
  return P(0,0) | P(1,8) | P(2,16) | P(3,24);
#undef P
}

unsigned short get16(const unsigned char *ptr)
{
#define P(i, n) (((unsigned short) ptr[i]) << n)
  return P(0,0) | P(1,8);
#undef P
}

static int xwrite(int fd, const void *buf, int len, const char *ofilename)
{
  int l = write(fd, buf, len);
  if (l < 0 || l != len)
  {
    fprintf(stderr, "%s: write error - disk full ?\n", ofilename);
    exit(1);
  }
  return l;
}

static int write_stub(int fd, const char *ofilename)
{
  static unsigned char *user_stub = NULL;
  static int user_stub_size = 0;

  /* no stub */
  if (opt_destubify)
    return 0;

  /* default stub */
  if (!opt_stubfile[0])
    return xwrite(fd, stub_bytes, sizeof(stub_bytes), ofilename);

  /* load stub from opt_stubfile[] */
  if (user_stub == NULL)
  {
    int fd = open(opt_stubfile, O_RDONLY|O_BINARY);
    if (fd >= 0)
    {
      user_stub_size = lseek(fd, 0, SEEK_END);
      if (user_stub_size > 0)
        user_stub = malloc(user_stub_size);
      if (user_stub != NULL)
      {
        if (lseek(fd, 0, SEEK_SET) != 0 ||
            read(fd, user_stub, user_stub_size) != user_stub_size)
        {
          free(user_stub);
          user_stub = NULL;
        }
      }
      close(fd);
    }
    if (user_stub == NULL)
    {
      fprintf(stderr, "stubify: error loading stub from file %s\n", opt_stubfile);
      exit(1);
    }
  }
  return xwrite(fd, user_stub, user_stub_size, ofilename);
}

int coff2exe(const char * const fname)
{
  char ifilename[PATH_MAX+1];
  char ofilename[PATH_MAX+1];
  int ifile;
  int ofile;
  char *ofname, *ofext;
  unsigned char filehdr_buf[20];
  unsigned char buf[4096];
  int rbytes, used_temp = 0, i, n;
  long coffset = 0;
  int max_file_size = 0;
  int coff_file_size = 0;
  int drop_last_four_bytes = 0;

  strcpy(ifilename, fname);
  ifile = open(ifilename, O_RDONLY|O_BINARY);
  if (ifile < 0)
  {
    perror(fname);
    return 1;
  }
  while (1)
  {
    lseek(ifile, coffset, 0);
    read(ifile, buf, 6);
    if (buf[0] == 'M' && buf[1] == 'Z') /* stubbed already, skip stub */
    {
      int blocks = get16(buf+4);
      int partial = get16(buf+2);
      coffset += blocks * 512;
      if (partial)
        coffset += partial - 512;
    }
    else if (buf[0] == 0x4c && buf[1] == 0x01) /* it's a COFF */
    {
      break;
    }
    else
    {
      fprintf(stderr, "Error: input file %s is not COFF or stubbed COFF\n", ifilename);
      close(ifile);
      return 1;
    }
  }

  strcpy(ofilename, fname);
  ofext = 0;
  for (ofname=ofilename; *ofname; ofname++)
  {
    if (strchr("/\\:", *ofname))
      ofext = 0;
    if (*ofname == '.')
      ofext = ofname;
  }
  if (ofext == 0)
    ofext = ofilename + strlen(ofilename);
  strcpy(ofext, opt_destubify ? ".cof" : ".exe");
  if (access(ofilename, 0) == 0)
    for (ofile=0; ofile<999; ofile++)
    {
      used_temp = 1;
      sprintf(ofext, ".%03d", ofile);
      if (access(ofilename, 0))
        break;
    }

  coff_file_size = lseek(ifile, 0, SEEK_END) - coffset;
  lseek(ifile, coffset, 0);

  read(ifile, filehdr_buf, 20); /* get the COFF header */
  lseek(ifile, get16(filehdr_buf+16), SEEK_CUR); /* skip optional header */
  n = get16(filehdr_buf + 2);
  for (i=0; i<n; i++) /* for each COFF section */
  {
    int addr, size, flags;
    unsigned char coffsec_buf[40];
    read(ifile, coffsec_buf, 40);
    size = get32(coffsec_buf+16);
    addr = get32(coffsec_buf+20);
    flags = get32(coffsec_buf+36);
    if (flags & 0x60) /* text or data */
    {
      int maybe_file_size = size + addr;
      if (max_file_size < maybe_file_size)
        max_file_size = maybe_file_size;
    }
  }
  if (coff_file_size == max_file_size + 4)
    /* it was built with "gcc -s" and has four too many bytes */
    drop_last_four_bytes = 1;

  lseek(ifile, coffset, 0);

  ofile = open(ofilename, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0777);
  if (ofile < 0)
  {
    perror(ofilename);
    return 1;
  }
  v_printf("stubify: %s -> %s", ifilename, ofilename);
  if (used_temp)
  {
    strcpy(ifilename, ofilename);
    strcpy(ofext, opt_destubify ? ".cof" : ".exe");
    v_printf(" -> %s", ofilename);
  }
  v_printf("\n");

  write_stub(ofile, ofilename);

#if defined(__DJGPP__)
  /* if 0 bytes are read (or an error occurs, the loop will be broken from */
  while (1) {
    __dpmi_regs r;
    int wb;
    /* bypass the normal read routine to avoid the unnecessary copying of the
     * file contents into extended memory (the data is not actually being
     * used, only copied from one file to another, and so easy access from gcc
     * compiled code is not needed).
     */
    r.x.ax = 0x3f00; /* dos read from file handle function */
    r.x.bx = ifile;
    r.x.cx = tbsize; /* number of bytes to read */
    r.x.dx = __tb & 15; /* transfer buffer offset */
    r.x.ds = __tb / 16; /* transfer buffer segment */
    __dpmi_int(0x21, &r);
    if (r.x.flags & 1)
      errno = __doserr_to_errno(r.x.ax);
    if ((rbytes=(r.x.flags & 1) ? -1 : r.x.ax) <= 0)
      break;
#else
  while ((rbytes=read(ifile, buf, sizeof(buf))) > 0)
  {
    int wb;
#endif

#if defined(__DJGPP__)
    /* bypass the normal write routine to avoid the unnecessary copying of the
     * file contents from extended memory.
     */
    r.x.ax = 0x4000; /* dos write to file handle function */
    r.x.bx = ofile;
    r.x.cx = rbytes; /* number of bytes to write */
    r.x.dx = __tb & 15; /* transfer buffer offset */
    r.x.ds = __tb / 16; /* transfer buffer segment */
    __dpmi_int(0x21, &r);
    wb = (r.x.flags & 1) ? -1 : r.x.ax;
    if (r.x.flags & 1)
      errno = __doserr_to_errno(r.x.ax);
#else
    wb = xwrite(ofile, buf, rbytes, ofilename);
#endif
    if (wb < 0)
    {
      perror(ofilename);
      close(ifile);
      close(ofile);
      unlink(ofilename);
      exit(1);
    }
    if (wb < rbytes)
    {
      fprintf(stderr, "%s: disk full\n", ofilename);
      close(ifile);
      close(ofile);
      unlink(ofilename);
      exit(1);
    }
  }

  if (drop_last_four_bytes)
  {
    off_t pos = lseek(ofile, -4, SEEK_END);
    if (pos <= 0 || ftruncate(ofile, pos) != 0)
    {
      perror(ofilename);
      close(ifile);
      close(ofile);
      unlink(ofilename);
      exit(1);
    }
  }

  close(ifile);
  close(ofile);

  if (used_temp)
  {
    unlink(ofilename);
#if defined(__DJGPP__)
    if (_rename(ifilename, ofilename))
#else
    if (rename(ifilename, ofilename))
#endif
    {
      fprintf(stderr, "rename of %s to %s failed.\n", ifilename, ofilename);
      perror("The error was");
      return 1;
    }
  }
  return 0;
}

void header(FILE *f)
{
  static int done = 0;
  if (done++ == 0)
    fprintf(f, "stubify 2.03-mfx2 for djgpp V2 executables, Copyright (C) 1995-2001 DJ Delorie\n");
}

void usage(void)
{
  FILE *f = stderr;
  header(f);
  fprintf(f, "\nUsage: stubify    [options] <program>...     stubify <program>\n");
  fprintf(f, "  or:  stubify -d [-v]      <program>...     destubify <program>\n");
  fprintf(f, "  or:  stubify -g [-v]      <file>           generate stub, write to <file>\n");
  fprintf(f, "Options:\n");
  fprintf(f, "  -v           be verbose\n");
  fprintf(f, "  -s <file>    use <file> instead of the builtin stub\n");
  fprintf(f, "\nProgram may be COFF or stubbed .exe, and may be COFF with .exe extension.\n");
  fprintf(f, "Resulting file will have .exe or .cof extension, according to operation mode.\n");

  fprintf(f, "\nThis program is NOT shareware or public domain.  It is copyrighted.\n");
  fprintf(f, "It is redistributable but only as part of a complete package.  If you\n");
  fprintf(f, "have a copy of this program, the place that you got it from is\n");
  fprintf(f, "responsible for making sure you are able to get it's sources as well.\n");

  exit(1);
}

int main(int argc, char **argv)
{
  int i;
  int exit_code = 0;

  assert(sizeof(stub_bytes) == 2048);

  for (i = 1; i < argc && argv[i][0] == '-'; i++)
  {
    if (strcmp(argv[i], "--") == 0)             /* end of options */
    {
      i++;
      break;
    }
    else if (strcmp(argv[i], "-d") == 0)
      opt_destubify = 1;
    else if (strcmp(argv[i], "-g") == 0)
      opt_generate = 1;
    else if (strcmp(argv[i], "-s") == 0)
    {
      if (++i >= argc)
        usage();
      strcpy(opt_stubfile, argv[i]);
    }
    else if (strcmp(argv[i], "-v") == 0)
      opt_verbose = 1;
    else
      usage();
  }
  if (i >= argc)
    usage();

  if (opt_generate)
  {
    char ofilename[PATH_MAX+1], *ofname, *ofext=NULL;
    int ofile;

    if (i + 1 != argc)          /* need exactly one argument */
      usage();
    strcpy(ofilename, argv[i]);
    for (ofname=ofilename; *ofname; ofname++)
    {
      if (strchr("/\\:", *ofname))
        ofext = 0;
      if (*ofname == '.')
        ofext = ofname;
    }
    if (ofext == NULL)
      ofext = ofilename + strlen(ofilename);
    strcpy(ofext, ".exe");

    if (opt_verbose)
      header(stdout);
    ofile = open(ofilename, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
    if (ofile < 0)
    {
      fprintf(stderr, "Cannot open output file to generate\n");
      perror(ofilename);
      return 1;
    }
    v_printf("stubify: generate stub file %s\n", ofilename);

    xwrite(ofile, stub_bytes, sizeof(stub_bytes), ofilename);
    close(ofile);
  }
  else
  {
    if (opt_verbose)
      header(stdout);
    while (i < argc)
      exit_code |= coff2exe(argv[i++]);
  }
  return exit_code;
}

